import napari
from  napari_segment_everything import segment_everything

viewer = napari.Viewer()

viewer.window.add_dock_widget(segment_everything.NapariSegmentEverything(viewer))

k=input("press close to exit") 



from segment_anything import sam_model_registry, SamPredictor 
from segment_anything.automatic_mask_generator import SamAutomaticMaskGenerator
import urllib.request
import warnings
from pathlib import Path
from typing import Optional
from skimage.measure import label                    
from skimage.measure import regionprops
from skimage import color

import toolz as tz
from napari.utils import progress

# Some code in this file copied from https://github.com/royerlab/napari-segment-anything/blob/main/src/napari_segment_anything/utils.py

SAM_WEIGHTS_URL = {
    "default": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth",
    "vit_h": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_h_4b8939.pth",
    "vit_l": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_l_0b3195.pth",
    "vit_b": "https://dl.fbaipublicfiles.com/segment_anything/sam_vit_b_01ec64.pth",
}


@tz.curry
def _report_hook(
    block_num: int,
    block_size: int,
    total_size: int,
    pbr: "progress" = None,
) -> None:
    downloaded = block_num * block_size
    percent = downloaded * 100 / total_size
    downloaded_mb = downloaded / 1024 / 1024
    total_size_mb = total_size / 1024 / 1024
    increment = int(percent) - pbr.n
    if increment > 1:  # faster than increment at every iteration
        pbr.update(increment)
    print(
        f"Download progress: {percent:.1f}% ({downloaded_mb:.1f}/{total_size_mb:.1f} MB)",
        end="\r",
    )


def download_weights(weight_url: str, weight_path: "Path"):
    print(f"Downloading {weight_url} to {weight_path} ...")
    pbr = progress(total=100)
    try:
        urllib.request.urlretrieve(
            weight_url, weight_path, reporthook=_report_hook(pbr=pbr)
        )
    except (
        urllib.error.HTTPError,
        urllib.error.URLError,
        urllib.error.ContentTooShortError,
    ) as e:
        warnings.warn(f"Error downloading {weight_url}: {e}")
        return None
    else:
        print("\rDownload complete.                            ")
    pbr.close()


def get_weights_path(model_type: str) -> Optional[Path]:
    """Returns the path to the weight of a given model architecture."""
    weight_url = SAM_WEIGHTS_URL[model_type]

    cache_dir = Path.home() / ".cache/tnia-sam"
    cache_dir.mkdir(parents=True, exist_ok=True)

    weight_path = cache_dir / weight_url.split("/")[-1]

    # Download the weights if they don't exist
    if not weight_path.exists():
        download_weights(weight_url, weight_path)

    return weight_path


def get_sam(model_type: str):
    sam = sam_model_registry[model_type](get_weights_path(model_type))
    #sam.to(self._device)
    return sam

def get_sam_predictor(model_type: str):
    sam = sam_model_registry[model_type](get_weights_path(model_type))
    #sam.to(self._device)
    return SamPredictor(sam)

def get_sam_automatic_mask_generator(model_type: str, 
                                     points_per_side=32, 
                                     pred_iou_thresh=0.1, 
                                     stability_score_thresh=0.1, 
                                     box_nms_thresh=0.5,
                                     crop_n_layers=1):
    
    sam = sam_model_registry[model_type](get_weights_path(model_type))
    sam_anything_predictor = SamAutomaticMaskGenerator(sam,
        points_per_side=int(points_per_side),
        #points_per_batch=64,
        pred_iou_thresh=pred_iou_thresh,
        stability_score_thresh=stability_score_thresh,
        #stability_score_offset=1.0
        box_nms_thresh=box_nms_thresh,
        crop_n_layers=crop_n_layers,
        #crop_nms_thresh=0.7,
        #crop_overlap_ratio: float = 512 / 1500,
        #crop_n_points_downscale_factor=1,
        #in_mask_region_area=0,
        #point_grids: Optional[List[np.ndarray]] = None,
        )
    #sam.to(self._device)
    return sam_anything_predictor

import numpy as np

def make_label_image_3d(masks):
    '''
    Creates a label image by adding one mask at a time onto an empty image
    '''
    num_masks = len(masks)
    label_image = np.zeros( (num_masks, masks[0]['segmentation'].shape[0], masks[0]['segmentation'].shape[1]), dtype='uint16')

    for enum, mask in enumerate(masks):
        mnarray = mask['segmentation']
        label_image[enum,:,:] = mnarray.astype('uint16') * (enum+1)

    return label_image

def filter_labels_3d_multi(label_image, sorted_results, stats, mins, maxes, napari_label=None): 
    for enum, result in enumerate(sorted_results):
        keep = all(min <= result[stat] <= max for stat, min, max in zip(stats, mins, maxes))
        if keep:
            if result['keep']==True:
                continue
            result['keep'] = True
            coords = np.where(result['segmentation'])
            if napari_label is None:
                temp = label_image[enum, :,:]
                temp[coords] = enum+1
            else:
                z = np.full(coords[0].shape, enum)
                coords = (z, coords[0], coords[1])
                if enum < label_image.shape[0]-1:
                    napari_label.data_setitem(coords, enum+1, True)
                else:
                    napari_label.data_setitem(coords, enum+1, True)
        else:
            if result['keep']==False:
                continue
            result['keep'] = False
            coords = np.where(result['segmentation'])
            if napari_label is None:
                temp = label_image[enum, :,:]
                temp[coords] = 0
            else:
                z = np.full(coords[0].shape, enum)
                coords = (z, coords[0], coords[1])
                if enum < label_image.shape[0]-1:
                    napari_label.data_setitem(coords, 0, True)
                else:
                    napari_label.data_setitem(coords, 0, True)

def add_properties_to_label_image(orig_image, sorted_results):

    hsv_image = color.rgb2hsv(orig_image)

    hue = 255*hsv_image[:,:,0]
    saturation = 255*hsv_image[:,:,1]
    intensity = 255*hsv_image[:,:,2]

    for enum, result in enumerate(sorted_results):
        segmentation = result['segmentation']
        coords = np.where(segmentation)
        regions = regionprops(segmentation.astype('uint8'))

        # calculate circularity
        result['circularity'] = 4*np.pi*regions[0].area / (regions[0].perimeter**2)
        result['solidity'] = regions[0].solidity
        intensity_pixels = intensity[coords]
        result['mean_intensity'] = np.mean(intensity_pixels)
        result['10th_percentile_intensity'] = np.percentile(intensity_pixels, 10)
        hue_pixels = hue[coords]
        result['mean_hue'] = np.mean(hue_pixels)
        saturation_pixels = saturation[coords]
        result['mean_saturation'] = np.mean(saturation_pixels)




# qypt imports
from qtpy.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QSpinBox,
    QDoubleSpinBox,
    QWidget,
    QSlider,
)

from qtpy.QtCore import Qt

class LabeledMinMaxSlider(QWidget):
    def __init__(self, text, min_value, max_value, initial_min_value, initial_max_value, tick_interval, stat, change_value_method, scale_slider = False):
        super().__init__()

        self.scale_slider = scale_slider

        self.label = QLabel(text)
        self.min_spinbox = QDoubleSpinBox()
        self.min_spinbox.setRange(min_value, max_value)
        self.min_spinbox.setValue(initial_min_value)
        self.min_spinbox.valueChanged.connect(self.change_value)
        self.min_spinbox.setKeyboardTracking(False)

        self.min_slider = QSlider(Qt.Horizontal)
        if scale_slider == False:
            self.min_slider.setMinimum(min_value)
            self.min_slider.setMaximum(max_value)
            self.min_slider.setValue(initial_min_value)
            self.min_slider.setTickPosition(QSlider.TicksBelow)
            self.min_slider.setTickInterval(tick_interval)
        else:
            self.min_slider.setMinimum(0)
            self.min_slider.setMaximum(100)
            self.min_slider.setValue(100 * int((initial_min_value - min_value) / (max_value - min_value)))
            self.min_slider.setTickPosition(QSlider.TicksBelow)
            self.min_slider.setTickInterval(1)
        self.min_slider.valueChanged.connect(self.change_value)

        self.max_spinbox = QDoubleSpinBox()
        self.max_spinbox.setRange(min_value, max_value)
        self.max_spinbox.setValue(initial_max_value)
        self.max_spinbox.valueChanged.connect(self.change_value)
        self.max_spinbox.setKeyboardTracking(False)

        self.max_slider = QSlider(Qt.Horizontal)
        
        if scale_slider == False:
            self.max_slider.setMinimum(min_value)
            self.max_slider.setMaximum(max_value)
            self.max_slider.setValue(initial_max_value)
            self.max_slider.setTickPosition(QSlider.TicksBelow)
            self.max_slider.setTickInterval(tick_interval)
        else:
            self.max_slider.setMinimum(0)
            self.max_slider.setMaximum(100)
            self.max_slider.setValue(100 * int((initial_max_value - min_value) / (max_value - min_value)))
            self.max_slider.setTickPosition(QSlider.TicksBelow)
            self.max_slider.setTickInterval(1)
        
        self.max_slider.valueChanged.connect(self.change_value)

        self.min_layout = QHBoxLayout()
        self.min_layout.setContentsMargins(5, 1, 5, 1)  # adjust the margins around the layout
        self.min_layout.addWidget(self.min_slider)
        self.min_layout.addWidget(self.min_spinbox)

        self.max_layout = QHBoxLayout()
        self.max_layout.setContentsMargins(5, 1, 5, 1)  # adjust the margins around the layout
        self.max_layout.addWidget(self.max_slider)
        self.max_layout.addWidget(self.max_spinbox)

        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(5, 1, 5, 1)  # adjust the margins around the layout
        # set spacing
        self.layout.setSpacing(0)
        
        #self.layout.addWidget(self.label)
        self.layout.addLayout(self.min_layout)
        self.layout.addLayout(self.max_layout)
        self.setLayout(self.layout)

        self.stat = stat
        self.change_value_method = change_value_method 

    def change_value(self):
        sender = self.sender()
        if sender in [self.min_slider, self.max_slider]:
            # Block signals from the spinboxes while changing their values
            self.min_spinbox.blockSignals(True)
            self.max_spinbox.blockSignals(True)
            if self.scale_slider == False:
                self.min_spinbox.setValue(self.min_slider.value())
                self.max_spinbox.setValue(self.max_slider.value())
            else:
                self.min_spinbox.setValue(self.min_slider.value() * (self.max_spinbox.maximum() - self.max_spinbox.minimum()) / 100 + self.max_spinbox.minimum())
                self.max_spinbox.setValue(self.max_slider.value() * (self.max_spinbox.maximum() - self.max_spinbox.minimum()) / 100 + self.max_spinbox.minimum())
            self.min_spinbox.blockSignals(False)
            self.max_spinbox.blockSignals(False)
        elif sender in [self.min_spinbox, self.max_spinbox]:
            # Block signals from the sliders while changing their values
            self.min_slider.blockSignals(True)
            self.max_slider.blockSignals(True)
            if self.scale_slider == False:
                self.min_slider.setValue(int(self.min_spinbox.value()))
                self.max_slider.setValue(int(self.max_spinbox.value()))
            else:
                self.min_slider.setValue(int(100 * (self.min_spinbox.value() - self.max_spinbox.minimum()) / (self.max_spinbox.maximum() - self.max_spinbox.minimum())))
                self.max_slider.setValue(int(100 * (self.max_spinbox.value() - self.max_spinbox.minimum()) / (self.max_spinbox.maximum() - self.max_spinbox.minimum())))
            self.min_slider.blockSignals(False)
            self.max_slider.blockSignals(False)

        self.change_value_method(self.stat, self.min_spinbox.value(), self.max_spinbox.value())

    def update_min_max(self, min_value, max_value):
        self.min_spinbox.setRange(min_value, max_value)
        self.max_spinbox.setRange(min_value, max_value)
        self.min_slider.setRange(min_value, max_value)
        self.max_slider.setRange(min_value, max_value)

class LabeledMinMaxSpinners(QWidget):

    def __init__(self, min_label_text, max_label_text, min_value, max_value, initial_min_value, initial_max_value, change_value_method):
        super().__init__()

        self.min_label = QLabel(min_label_text)
        self.min_spinbox = QDoubleSpinBox()
        self.min_spinbox.setRange(min_value, max_value)
        self.min_spinbox.setValue(initial_min_value)

        if change_value_method is not None:
            self.min_spinbox.valueChanged.connect(change_value_method)

        self.max_label = QLabel(max_label_text)
        self.max_spinbox = QDoubleSpinBox()
        self.max_spinbox.setRange(min_value, max_value)
        self.max_spinbox.setValue(initial_max_value)
        
        if change_value_method is not None:
            self.max_spinbox.valueChanged.connect(change_value_method)

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(10, 2, 10, 2)  # adjust the margins around the layout
        self.layout.addWidget(self.min_label)
        self.layout.addWidget(self.min_spinbox)
        self.layout.addWidget(self.max_label)
        self.layout.addWidget(self.max_spinbox)

        self.setLayout(self.layout)

class LabeledSpinner(QWidget):
    def __init__(self, label_text, min_value, max_value, default_value, change_value_method, is_double=False):
        super().__init__()

        self.label = QLabel(label_text)
        self.spinner = QDoubleSpinBox() if is_double else QSpinBox()
        self.spinner.setRange(min_value, max_value)
        self.spinner.setValue(default_value)

        if change_value_method is not None:
            self.spinner.valueChanged.connect(change_value_method)

        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(10, 2, 10, 2)  # adjust the margins around the layout
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.spinner)

        self.setLayout(self.layout)



from magicgui.widgets import create_widget
from napari.layers import Image
import numpy as np
from skimage import color, util
from typing import Optional

from napari_segment_everything.widgets import LabeledSpinner, LabeledMinMaxSlider
from napari_segment_everything.sam_helper import get_sam_automatic_mask_generator, make_label_image_3d, add_properties_to_label_image, filter_labels_3d_multi
import pickle

# qypt improts
from qtpy.QtWidgets import (
    QVBoxLayout,
    QWidget,
    QPushButton,
    QComboBox,
    QLabel,
    QHBoxLayout,
    QFileDialog,
    QStackedWidget,
    QGroupBox,
    QSizePolicy,
    QMessageBox
)

class NapariSegmentEverything(QWidget):

    def __init__(self, napari_viewer, parent=None):

        QWidget.__init__(self, parent=parent)
        self.viewer = napari_viewer

        self.results = None
        self.block_stats = False
        
        self.initUI()

        self._3D_labels_layer = self.viewer.add_labels(
            data=np.zeros((256, 256, 256), dtype=int),
            name="SAM 3D labels",
        )

        self._3D_labels_layer.mouse_double_click_callbacks.append(self.test_double_click)
        self.load_image(self.im_layer_widget.value)

        self.msg = QMessageBox()

    def test_double_click(self, layer, event):
        index = int(layer._value)-1  

        result = self.results[index]
        stats = [
            f"Area: {result['area']}",
            f"Circularity: {result['circularity']:.2f}",
            f"Solidity: {result['solidity']:.2f}",
            f"Mean intensity: {result['mean_intensity']:.2f}",
            f"10th percentile intensity: {result['10th_percentile_intensity']:.2f}",
            f"Mean hue: {result['mean_hue']:.2f}",
            f"Mean saturation: {result['mean_saturation']:.2f}",
            f"Predicted IOU: {result['predicted_iou']:.2f}",
            f"Stability score: {result['stability_score']:.2f}",
        ]
        stats_text = "\n".join(stats)

        self.msg.setIcon(QMessageBox.Information)
        self.msg.setText("Result Stats")
        self.msg.setInformativeText(stats_text)
        self.msg.setWindowTitle("Result Stats")
        self.msg.setModal(True)
        self.msg.show()

    def initUI(self):
        
        self.setWindowTitle('Segment Everything')
        layout = QVBoxLayout()
        layout.setSpacing(30)

        ##### 1.  SAM Parameters ####

        self.sam_parameters_group = QGroupBox("1. Generate 3D labels")
        self.sam_layout = QVBoxLayout()
        self.sam_parameters_group.setLayout(self.sam_layout)

        # add open results button
        self.open_project_button = QPushButton("Open previous...")
        self.open_project_button.clicked.connect(self.open_project)
        self.sam_layout.addWidget(self.open_project_button)

        # Dropdown for selecting the model
        model_label = QLabel("Select Model")# Dropdown for selecting the model
        self.model_dropdown = QComboBox()
        self.model_dropdown.addItems(["vit_b"])
        model_layout = QHBoxLayout()
        model_layout.addWidget(model_label)
        model_layout.addWidget(self.model_dropdown)
        self.sam_layout.addLayout(model_layout)

        # control for selecting the image
        image_layout = QHBoxLayout()
        image_label = QLabel("Select Image")
        self.im_layer_widget = create_widget(annotation=Image, label="Image:")
        self.im_layer_widget.changed.connect(self.load_image)
        image_layout.addWidget(image_label)
        image_layout.addWidget(self.im_layer_widget.native)
        self.sam_layout.addLayout(image_layout)
        
        self.im_layer_widget.reset_choices()
        self.viewer.layers.events.inserted.connect(self.im_layer_widget.reset_choices)
        self.viewer.layers.events.removed.connect(self.im_layer_widget.reset_choices)

        self.points_per_side_spinner = LabeledSpinner("Points per side", 4, 100, 32, None)
        self.pred_iou_thresh_spinner = LabeledSpinner("Pred IOU Threshold", 0, 1, 0.1, None, is_double=True)
        self.stability_score_thresh_spinner = LabeledSpinner("Stability Score Threshold", 0, 1, 0.1, None, is_double=True)
        self.box_nms_thresh_spinner = LabeledSpinner("Box NMS Threshold", 0, 1, 0.7, None, is_double=True)
        self.crop_n_layers_spinner = LabeledSpinner("Crop N Layers", 0, 10, 0, None)

        self.sam_layout.addWidget(self.points_per_side_spinner)
        self.sam_layout.addWidget(self.pred_iou_thresh_spinner)
        self.sam_layout.addWidget(self.stability_score_thresh_spinner)
        self.sam_layout.addWidget(self.box_nms_thresh_spinner)
        self.sam_layout.addWidget(self.crop_n_layers_spinner)

        # add process button
        self.process_button = QPushButton("Generate 3D labels")
        self.process_button.clicked.connect(self.process)
        self.sam_layout.addWidget(self.process_button)

        layout.addWidget(self.sam_parameters_group)

        ##### 2.  Filter Results #####

        self.filter_results_group = QGroupBox("2. Filter 3D Labels")
        self.filter_layout = QVBoxLayout()
        self.filter_results_group.setLayout(self.filter_layout)

        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.filter_results_group.setSizePolicy(sizePolicy)

        self.min_max_area_slider = LabeledMinMaxSlider("Area", 0, 1000000, 0, 1000000, 1000, 'area', self.change_stat)
        self.min_max_label_num_slider = LabeledMinMaxSlider("Label Num", 0, 500, 0, 500, 1, 'label_num', self.change_stat)
        self.min_max_solidity_slider = LabeledMinMaxSlider("Solidity", 0, 1, 0, 1, .01, 'solidity', self.change_stat, True)
        self.min_max_circularity_slider = LabeledMinMaxSlider("Circularity", 0, 1, 0, 1, .01, 'circularity', self.change_stat, True)
        self.min_max_mean_intensity_slider = LabeledMinMaxSlider("Mean Intensity", 0, 255, 0, 255, 10, 'mean_intensity', self.change_stat)
        self.min_max_p10_intensity_slider = LabeledMinMaxSlider("10th Percentile Intensity", 0, 1000, 0, 1000, 10, '10th_percentile_intensity', self.change_stat)
        self.min_max_hue_slider = LabeledMinMaxSlider("Hue", 0, 255, 0, 255, 10, 'hue', self.change_stat)
        self.min_max_saturation_slider = LabeledMinMaxSlider("Saturation", 0, 255, 0, 255, 10, 'saturation', self.change_stat)
        self.min_max_iou_slider = LabeledMinMaxSlider("IOU", 0, 1, 0, 1, .01, 'iou', self.change_stat, True)
        self.min_max_stability_score_slider = LabeledMinMaxSlider("Stability", 0, 1, 0, 1, .01, 'stability_score', self.change_stat, True)

        self.sliders = [self.min_max_area_slider, self.min_max_label_num_slider, self.min_max_solidity_slider, self.min_max_circularity_slider, self.min_max_mean_intensity_slider, self.min_max_p10_intensity_slider, self.min_max_hue_slider, self.min_max_saturation_slider, self.min_max_iou_slider, self.min_max_stability_score_slider]

        self.combo_box = QComboBox()
        self.combo_box.addItems(["Area", "Label Num", "Solidity", "Circularity", "Mean Intensity", "10th Percentile Intensity", "Hue", "Saturation", "IOU", "Stability Score"])
        self.combo_box.currentIndexChanged.connect(self.change_slider)

        self.stacked_widget = QStackedWidget()
        for slider in self.sliders:
            self.stacked_widget.addWidget(slider)

        select_slider_layout = QHBoxLayout()
        select_slider_layout.addWidget(QLabel("Select Stat:"))
        select_slider_layout.addWidget(self.combo_box)
        self.filter_layout.addLayout(select_slider_layout)
        self.stacked_widget.setFixedHeight(100)
        self.filter_results_group.setFixedHeight(160)
        self.filter_layout.addWidget(self.stacked_widget)        
        
        layout.addWidget(self.filter_results_group)

        self.make_2d_labels_groups = QGroupBox("3. Generate 2D Labels")
        self.make_2d_labels_layout = QVBoxLayout()
        self.make_2d_labels_groups.setLayout(self.make_2d_labels_layout)
        
        # add 2d labels options
        self.labels_options_layout = QHBoxLayout()
        self.labels_options_label = QLabel("2D Labels Options")
        self.labels_options_combo = QComboBox()
        self.labels_options_combo.addItems(["Small in front", "Big in front"])
        self.labels_options_layout.addWidget(self.labels_options_label)
        self.labels_options_layout.addWidget(self.labels_options_combo)
        
        # add make 2d labels button
        self.make_2d_labels_button = QPushButton("Make 2D Labels")
        self.make_2d_labels_button.clicked.connect(self.make_2d_labels)

        self.make_2d_labels_layout.addLayout(self.labels_options_layout)
        self.make_2d_labels_layout.addWidget(self.make_2d_labels_button)

        layout.addWidget(self.make_2d_labels_groups)

        # add save results button
        self.save_project_button = QPushButton("Save SAM project")
        self.save_project_button.clicked.connect(self.save_project)
        layout.addWidget(self.save_project_button)
        
        self.setLayout(layout)

    def change_slider(self, index):
        self.stacked_widget.setCurrentIndex(index)

    def open_project(self):
        options = QFileDialog.Options()
        options |= QFileDialog.ReadOnly
        file_name, _ = QFileDialog.getOpenFileName(self, "QFileDialog.getOpenFileName()", "", "Pickle Files (*.pkl)", options=options)
        if file_name:
            with open(file_name, 'rb') as f:
                project = pickle.load(f)
                self.results = project['results']
                image = project['image']
                self.viewer.add_image(image)
                self.image = image
        
        self._3D_labels_layer.data = make_label_image_3d(self.results)
        self.viewer.dims.ndisplay = 3

        add_properties_to_label_image(self.image, self.results)
        self.update_slider_min_max() 

    def save_project(self):
        options = QFileDialog.Options()
        file_name, _ = QFileDialog.getSaveFileName(self, "QFileDialog.getSaveFileName()", "", "Pickle Files (*.pkl)", options=options)
        if file_name:
            project = {'results': self.results, 'image': self.image}
            with open(file_name, 'wb') as f:
                pickle.dump(project, f)

    def process(self):
        if self.image is None:
            return
        
        points_per_side = self.points_per_side_spinner.spinner.value()
        pred_iou_thresh = self.pred_iou_thresh_spinner.spinner.value()
        stability_score_thresh = self.stability_score_thresh_spinner.spinner.value()
        box_nms_thresh = self.box_nms_thresh_spinner.spinner.value()
        crop_n_layers = self.crop_n_layers_spinner.spinner.value()
        
        self._predictor = get_sam_automatic_mask_generator("vit_b", points_per_side=points_per_side, 
                                                           pred_iou_thresh=pred_iou_thresh, 
                                                           stability_score_thresh=stability_score_thresh, 
                                                           box_nms_thresh=box_nms_thresh,
                                                           crop_n_layers=crop_n_layers
                                                           )
        
        self.results = self._predictor.generate(self.image)
        
        self.results = sorted(self.results, key=lambda x: x['area'], reverse=False)

        print(len(self.results), 'objects found')
        
        label_num = 1
        for result in self.results:
            result['keep'] = True
            result['label_num'] = label_num
            label_num += 1

        add_properties_to_label_image(self.image, self.results)

        label_image = make_label_image_3d(self.results)

        print(label_image.shape)

        self.update_slider_min_max() 
        self._3D_labels_layer.data = label_image
        self.viewer.dims.ndisplay = 3

    def update_slider_min_max(self):

        stats = ['area', 'label_num', 'solidity', 'circularity', 'mean_intensity', '10th_percentile_intensity', 'mean_hue', 'mean_saturation', 'predicted_iou', 'stability_score']
        for stat in stats:
            min_ = min([result[stat] for result in self.results])
            max_ = max([result[stat] for result in self.results])

            print(stat, min_, max_)

        num_labels = len(self.results)

        self.block_stats = True
        self.min_max_label_num_slider.min_spinbox.setRange(0, num_labels)
        self.min_max_label_num_slider.max_spinbox.setRange(0, num_labels)
        self.min_max_label_num_slider.min_slider.setRange(0, num_labels)
        self.min_max_label_num_slider.max_slider.setRange(0, num_labels)
        #self.min_max_label_num_slider.min_slider.setValue(0)
        self.min_max_label_num_slider.max_slider.setValue(num_labels)
        self.block_stats = False
        
    def change_stat(self, stat, min_value, max_value):
        if (self.block_stats == True):
            return
        
        if (self.viewer.dims.ndisplay != 3):
            return
        
        if (self.results is None):
            return
        
        # max of predicted_iou stat
        min_iou = min([result['predicted_iou'] for result in self.results])
        max_iou = max([result['predicted_iou'] for result in self.results])

        min_stability_score = min([result['stability_score'] for result in self.results])
        max_stability_score = max([result['stability_score'] for result in self.results])

        min_iou = self.min_max_iou_slider.min_spinbox.value()
        max_iou = self.min_max_iou_slider.max_spinbox.value()

        min_stability_score = self.min_max_stability_score_slider.min_spinbox.value()
        max_stability_score = self.min_max_stability_score_slider.max_spinbox.value()

        min_circularity = self.min_max_circularity_slider.min_spinbox.value()
        max_circularity = self.min_max_circularity_slider.max_spinbox.value()

        # TODO: change below to a dictionary to make it easier to add new stats

        stats = ['area', 'label_num', 'solidity', 'circularity', 'mean_intensity', '10th_percentile_intensity', 'mean_hue', 'mean_saturation', 'predicted_iou', 'stability_score']
        mins = [self.min_max_area_slider.min_spinbox.value(), self.min_max_label_num_slider.min_spinbox.value(), 
                self.min_max_solidity_slider.min_spinbox.value(), min_circularity, self.min_max_mean_intensity_slider.min_spinbox.value(), 
                self.min_max_p10_intensity_slider.min_spinbox.value(), self.min_max_hue_slider.min_spinbox.value(), self.min_max_saturation_slider.min_spinbox.value(),
                min_iou, min_stability_score]
        maxs = [self.min_max_area_slider.max_spinbox.value(), self.min_max_label_num_slider.max_spinbox.value(), 
                self.min_max_solidity_slider.max_spinbox.value(), max_circularity, self.min_max_mean_intensity_slider.max_spinbox.value(), 
                self.min_max_p10_intensity_slider.max_spinbox.value(), self.min_max_hue_slider.max_spinbox.value(), self.min_max_saturation_slider.max_spinbox.value(),
                max_iou, max_stability_score]
        if self.results is not None:
            filter_labels_3d_multi(self._3D_labels_layer.data, self.results, stats, mins, maxs, napari_label=self._3D_labels_layer)
     
    def refresh_slider(self):
        area_min, area_max = 0, np.inf
        label_num_min, label_num_max = 0, np.inf
        solidity_min, solidity_max = 0, 1
        mean_intensity_min, mean_intensity_max = 0, np.inf
        p10_intensity_min, p10_intensity_max = 0, np.inf

        for result in self.results:
            area_min = min(area_min, result['area'])
            area_max = max(area_max, result['area'])
            label_num_min = min(label_num_min, result['label_num'])
            label_num_max = max(label_num_max, result['label_num'])
            solidity_min = min(solidity_min, result['solidity'])
            solidity_max = max(solidity_max, result['solidity'])
            mean_intensity_min = min(mean_intensity_min, result['mean_intensity'])
            mean_intensity_max = max(mean_intensity_max, result['mean_intensity'])
            p10_intensity_min = min(p10_intensity_min, result['10th_percentile_intensity'])
            p10_intensity_max = max(p10_intensity_max, result['10th_percentile_intensity'])

    def make_2d_labels(self):
        min_max = self.labels_options_combo.currentText()
        
        label_image = self._3D_labels_layer.data
        
        if min_max == "Small in front":
             # Create a masked array where zeros are masked
            masked_label_image = np.ma.masked_equal(label_image, 0)
            # Perform the min projection on the masked array
            labels = np.ma.min(masked_label_image, axis=0).filled(0)
        else:
            labels = np.max(label_image, axis=0)
        self.viewer.add_labels(labels, name="2D labels")
     
    def load_image(self, im_layer: Optional[Image]) -> None:
        
        if im_layer is None:
            return

        if im_layer.ndim != 2:
            raise ValueError(
                f"Only 2D images supported. Got {im_layer.ndim}-dim image."
            )

        image = im_layer.data
        if not im_layer.rgb:
            image = color.gray2rgb(image)

        elif image.shape[-1] == 4:
            # images with alpha
            image = color.rgba2rgb(image)

        if np.issubdtype(image.dtype, np.floating):
            image = image - image.min()
            image = image / image.max()

        self.image = util.img_as_ubyte(image)

        #self._mask_layer.data = np.zeros(self.image.shape[:2], dtype=int)
        #self._labels_layer.data = np.zeros(self.image.shape[:2], dtype=int)

        max_area = image.shape[0] * image.shape[1]

        self.min_max_area_slider.min_spinbox.setRange(0, max_area)
        self.min_max_area_slider.max_spinbox.setRange(0, max_area)
        self.min_max_area_slider.min_slider.setRange(0, max_area)
        self.min_max_area_slider.max_slider.setRange(0, max_area)

        for i in range(len(self.viewer.layers)):
            if self.viewer.layers[i].name == "SAM 3D labels":
                self.viewer.layers.move(i, -1)
                break

        

